import numpy as np

class Node:
    def __init__(self, **kwargs):
        self.val = 0
        self.children = [] 
        self.grad_value = None #the adjoint
        
        # QUESTION - HOW DO WE HANDLE CHILDREN IN kwargs?
        if 'val' in kwargs: 
            self.val = kwargs['val']
            
    ### KEY TO THE REVERSE MODE: ###
    # The recursive differentiation function 
    # Source (Adapted From): https://rufflewind.com/2016-12-30/reverse-mode-automatic-differentiation
    def grad(self):
        # recurse only if the value is not yet cached
        if self.grad_value is None:
            # recursively calculate derivative (adjoint) using chain rule
            self.grad_value = sum(weight * node.grad()
                                  for weight, node in self.children)
        return self.grad_value

    ### MATHEMATICAL OPERATORS ###
    # 1) Multiplication, 2) Division, 3) Addition, 4) Subtraction, 5) Power
    
    # 1) Multiplication
    def __mul__(self, other):
        out = Node()
        try:
            out.val = self.val*other.val
            self.children.append((other.val, out)) # weight = ∂z/∂self = other.value
            other.children.append((self.val, out)) # weight = ∂z/∂other = self.value
        except AttributeError:
            try:
                out.val = self.val*other
                self.children.append((other, out)) # weight = ∂z/∂self = other.value
            except Exception as e:
                raise TypeError('Multiplication failed: `other` not an `Node` object neither a real number.')
        return out

    def __rmul__(self, other):
        return self.__mul__(other)

    # 2) Division
    def __truediv__(self, other):
        out = Node()
        try:
            out.val = self.val / other.val
            self.children.append((1/other.val, out)) # weight = ∂z/∂self = 1/other.value
            other.children.append((-1*self.val/other.val**2, out)) # weight = ∂z/∂other = -1*self.value/ other.value**2
        except AttributeError:
            try:
                out.val = self.val/other
                self.children.append((1/other, out)) # weight = ∂z/∂self = 1/other
            except Exception as e:
                raise TypeError('Division failed: `other` not an `Node` object neither a real number.')
        return out
    
    def __rtruediv__(self, other):
        out = Node()
        
        try:
            out.val = other.val / self.val
            self.children.append((-1*other.val/ self.val**2, out)) # weight = ∂z/∂self = -1*other.value/ self.value**2
            other.children.append((1/self.val, out)) # weight = ∂z/∂other = 1 / self.value
        except AttributeError:
            try:
                out.val = other / self.val
                self.children.append((-1*other/ self.val**2, out)) # weight = ∂z/∂self = -1*other/ self.value**2
            except Exception as e:
                raise TypeError('Reverse Division failed: `self` not an `Node` object neither a real number.')
        
        return out
        
    # 3) Addition
    def __add__(self, other):
        out = Node()
        try:
            out.val = self.val + other.val
            self.children.append((1.0, out)) # weight = ∂z/∂self = 1
            other.children.append((1.0, out)) # weight = ∂z/∂other = 1            
        except AttributeError:
            try:
                out.val = self.val + other
                self.children.append((1.0, out)) # weight = ∂z/∂self = 1
            except Exception as e:
                raise TypeError('Addition failed: `other` not an `Node` object neither a real number.')
        return out
    
    def __radd__(self, other):
        return self.__add__(other)  
    
    # 4) Subtraction
    def __sub__(self, other):
        out = Node()
        try:
            out.val = self.val - other.val
            self.children.append((1.0, out)) # weight = ∂z/∂self = 1
            other.children.append((-1.0, out)) # weight = ∂z/∂other = -1            
        except AttributeError:
            try:
                out.val = self.val - other
                self.children.append((1.0, out)) # weight = ∂z/∂self = 1
            except Exception as e:
                raise TypeError('Subtraction failed: `other` not an `Node` object neither a real number.')
        return out
    
    def __rsub__(self, other):
        out = Node()
        try:
            out.val = other.val - self.val
            self.children.append((-1.0, out)) # weight = ∂z/∂self = 1
            other.children.append((1.0, out)) # weight = ∂z/∂other = -1            
        except AttributeError:
            try:
                out.val = other - self.val
                self.children.append((-1.0, out)) # weight = ∂z/∂self = 1
            except Exception as e:
                raise TypeError('Reverse Subtraction failed: `self` not an `Node` object neither a real number.')
        return out
    
    # 5) Power
    #NOTE left argument (self) MUST BE > 0
    def __pow__(self, other):    
        out = Node()
        try:
            out.val = self.val**other.val
            self.children.append((other.val * self.val ** (other.val - 1), out)) # weight = ∂z/∂self = other.value * self.value ** (other.value - 1)
            other.children.append((self.val ** other.val * np.log(self.val), out)) # weight = ∂z/∂other = self.value ** other.value * log(self.value)
        except AttributeError:
            try:
                out.val = self.val**other
                self.children.append((other * self.val ** (other - 1), out)) # weight = ∂z/∂self = other * self.value ** (other - 1)
            except Exception as e:
                raise TypeError('Power failed: `other` not an `Node` object neither a real number.')
        return out

    def __rpow__(self, other):    
        out = Node()
        try:
            out.val = other.val ** self.val
            self.children.append((other.val ** self.val * np.log(other.val), out)) # weight = ∂z/∂self = self.value ** other.value * log(self.value)
            other.children.append((self.val * other.val ** (self.val - 1), out)) # weight = ∂z/∂other = other.value * self.value ** (other.value - 1)
        except AttributeError:
            try:
                out.val = other ** self.val
                self.children.append((other ** self.val * np.log(other), out)) # weight = ∂z/∂self = self.value ** other.value * log(self.value)
            except Exception as e:
                raise TypeError('Reverse Power failed: `self` not an `Node` object neither a real number.')
        return out
    
    ### CLASS METHODS ###
    # 1) len, 2) str, 3) repr, etc.
    # TO DO
    
### NUMPY FUNCTIONS ###
# 1) log (i.e. ln), 2) exp, 3) sin, 4) cos, 5) tan

# 1) Logarithm    
def log(a):
    out = Node()
    try:
        out.val = np.log(a.val)
        a.children.append((1/a.val, out)) # weight = ∂z/∂self = 1 / self
    except AttributeError:
        try:
            out.val = np.log(a)
        except Exception as e:
            raise TypeError('Logarithm failed: `input` not an `Node` object, positive number or a real number.')
    return out

# 2) Exponent
def exp(a):
    out = Node()
    try:
        out.val = np.exp(a.val)
        a.children.append((np.exp(a.val), out)) # weight = ∂z/∂self = exp(self)
    except AttributeError:
        try:
            out.val = np.exp(a)
        except Exception as e:
            raise TypeError('Exponential failed: `input` not an `Node` object or a real number.')
    return out

# 3) Sin        
def sin(a):
    out = Node()
    try:
        out.val = np.sin(a.val)
        a.children.append((np.cos(a.val), out)) # weight = ∂z/∂self = cos(self)
    except AttributeError:
        try:
            out.val = np.sin(a)
        except Exception as e:
            raise TypeError('Sin failed: `input` not an `Node` object or a real number.')
    return out

# 4) Cos        
def cos(a):
    out = Node()
    try:
        out.val = np.cos(a.val)
        a.children.append((-1*np.sin(a.val), out)) # weight = ∂z/∂self = -sin(self)
    except AttributeError:
        try:
            out.val = np.cos(a)
        except Exception as e:
            raise TypeError('Cos failed: `input` not an `Node` object or a real number.')
    return out

# 5) Tan        
def tan(a):
    out = Node()
    try:
        out.val = np.tan(a.val)
        a.children.append((1/(np.cos(a.val))**2, out)) # weight = ∂z/∂self = 1/(cos(self))^2
    except AttributeError:
        try:
            out.val = np.tan(a)
        except Exception as e:
            raise TypeError('Tan failed: `input` not an `Node` object or a real number.')
    return out



import numpy as np

class Variable:
    def __init__(self,**kwargs):
        self.val = 0
        self.der = 1    
        
        if 'val' in kwargs:
            self.val = kwargs['val']
        if 'der' in kwargs:
            self.der = kwargs['der']

    def __mul__(self, other):
        out = Variable()
        try:
            out.val = self.val*other.val
            out.der = self.val*other.der + self.der*other.val
        except AttributeError:
            try:
                out.val = self.val*other
                out.der = other*self.der
            except Exception as e:
                # print(e)
                raise TypeError('Multiplication failed: `other` not an `AutoDiffToy` neither a real number.')
        return out

    def __rmul__(self, other):
        return self.__mul__(other)

    def __add__(self, other):
        out = Variable()
        try:
            out.val = self.val + other.val
            out.der = self.der + other.der
        except AttributeError:
            try:
                out.val = self.val + other
                out.der = self.der
            except Exception as e:
                #print(e)
                raise TypeError('Addition failed: `other` not an `AutoDiffToy` neither a real number.')
        return out
    
    def __radd__(self, other):
        return self.__add__(other)  

    def __sub__(self, other):
        out = Variable()
        try:
            out.val = self.val - other.val
            out.der = self.der - other.der
        except AttributeError:
            try:
                out.val = self.val - other
                out.der = self.der
            except Exception as e:
                # print(e)
                raise TypeError('Subtraction failed: `other` is neither an `AutoDiffToy` nor a real number.')
        return out
    
    def __rsub__(self, other):
        out = Variable()
        try:
            out.val = other.val - self.val
            out.der = other.der - self.der
        except AttributeError:
            try:
                out.val = other - self.val
                out.der = -self.der
            except Exception as e:
                # print(e)
                raise TypeError('Subtraction failed: `other` is neither an `AutoDiffToy` nor a real number.')
        return out

    def __pow__(self, other):
        out = Variable()
        try:
            out.val = self.val ** other.val
            out.der = (np.log(self.val) + 1) * (self.val ** other.val)
        except Exception as e:
            try:
                out.val = self.val**other
                out.der = other*(self.val**(other-1))*self.der
            except Exception as e:
                raise TypeError('Power raised to the base value failed: `other` is neither an `AutoDiffToy` nor a real number.')
        return out

    def __truediv__(self, other):
        out = Variable()
        try:
            out.val = self.val / other.val
            out.der = (self.der - (self.val / other.val) * other.der) / other.val
        except AttributeError:
            try:
                out.val = self.val / other
                out.der = self.der
            except Exception as e:
                # print(e)
                raise TypeError('Multiplication failed: `other` not an `AutoDiffToy` neither a real number.')
        return out

    def __rtruediv__(self, other):
        out = Variable()
        try:
            out.val = other.val / self.val
            out.der = (other.der - (other.val / self.val) * self.der) / self.val
        except AttributeError:
            try:
                out.val = other / self.val
                out.der = self.der
            except Exception as e:
                # print(e)
                raise TypeError('Multiplication failed: `other` not an `AutoDiffToy` neither a real number.')
        return out

    def sin(a):
        out = Variable()
        out.val = np.sin(a.val)
        out.der = np.cos(a.val)*a.der
        return out

    def cos(a):
        out = Variable()
        out.val = np.cos(a.val)
        out.der = -np.sin(a.val)*a.der
        return out

    def tanh(a):
        out = Variable()
        out.val = np.tanh(a.val)
        out.der = (1 - np.tanh(a.val)**2)*a.der
        return b

    def exp(a):
        out = Variable()
        out.val = np.exp(a.val)
        out.der = np.exp(a.val)*a.der
        return out

    def log(a):
        out = Variable()
        out.val = np.log(a.val)
        out.der = (1/a.val)*a.der
        return out



import numpy as np
from forward import Variable

if __name__=='__main__':

    # Forward, single function output
    # Expected (f1.val == -0.172, f1.der == 3.94)

    x1 = Variable(val=np.pi/16, der=1)
    f1 = x1 - np.exp(-2*(np.sin(4*x1)**2))


    print(f1.val, f1.der)



import reverse as rv
from reverse import Node

if __name__=='__main__':

    x = Node(0.5)
    y = Node(4.2)

    a = x * y
    a.grad_value = 1.0
    print("∂a/∂x = {}".format(x.grad())) # ∂a/∂x = 4.2

    # NOTE THE OUTPUT FUNCTION CAN'T HAVE THE SAME VARIABLE NAME AS OTHERS
    x = Node(2)
    y = Node(1)

    b = x * x * y * y
    b.grad_value = 1.0
    print("∂a/∂x = {}".format(x.grad())) # ∂a/∂x = 4.0



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 16 15:19:53 2020

@author: david
"""

## Usage case

import reverse as rv
from reverse import Node

if __name__=='__main__':

    # NOTE THE OUTPUT FUNCTION CAN'T HAVE THE SAME VARIABLE NAME AS OTHER FUNCTIONS

    # MULTIPLICATION TESTS
    x = Node(val = 0.5)
    y = Node(val = 4.2)

    a = x * y
    a.grad_value = 1.0
    print("∂a/∂x = {}".format(x.grad())) # ∂a/∂x = 4.2

    x = Node(val = 2)
    y = Node(val = 1)

    b = x * x * y * y
    b.grad_value = 1.0
    print("∂b/∂x = {}".format(x.grad())) # ∂b/∂x = 4.0

    # ADDITION TEST
    x = Node(val = 2)
    y = Node(val = 1)

    c = x + x + y
    c.grad_value = 1.0
    print("∂c/∂x = {}".format(x.grad())) # ∂c/∂x = 2.0

    x = Node(val = 2)

    d = 1.0 + x + x
    d.grad_value = 1.0
    print("∂d/∂x = {}".format(x.grad())) # ∂d/∂x = 2.0
    print(d.val)

    # SUBTRACTION TEST
    x = Node(val = 2)
    y = Node(val = 1)

    e = 1 - x - x - y
    e.grad_value = 1.0
    print("∂e/∂x = {}".format(x.grad())) # ∂e/∂x = -2
    print(e.val) # -4

    x = Node(val = 2)
    y = Node(val = 1)

    f = x - y
    f.grad_value = 1.0
    print("∂f/∂x = {}".format(y.grad())) # ∂f/∂x = -1
    print(f.val) # 1

    # DIVISION TESTS
    x = Node(val = 0.5)
    y = Node(val = 4.2)

    g = x / y
    g.grad_value = 1.0
    print("∂g/∂y = {}".format(y.grad())) # ∂g/∂y = -0.028

    y = Node(val = 4.2)

    h = 0.5 / y
    h.grad_value = 1.0
    print("∂h/∂y = {}".format(y.grad())) # ∂h/∂y = -0.028

    # POWER TESTS
    x = Node(val = 1.5)
    y = Node(val = 2.5)

    i = x ** y #NOTE x MUST BE > 0
    i.grad_value = 1.0
    print("∂i/∂x = {}".format(x.grad())) # ∂i/∂x = 4.593
    print("∂i/∂y = {}".format(y.grad())) # ∂i/∂y = 1.117

    y = Node(val = 2.5)

    j = 1.5 ** y
    j.grad_value = 1.0
    print("∂j/∂y = {}".format(y.grad())) # ∂j/∂y = 1.117

    x = Node(val = 1.5)

    k = x ** x
    k.grad_value = 1.0
    print("∂k/∂y = {}".format(x.grad())) # ∂k/∂x = 2.582

    x = Node(val = 1.5)

    k = x ** x
    k.grad_value = 1.0
    print("∂k/∂y = {}".format(x.grad())) # ∂k/∂x = 2.582


    # LOGARITHM TESTS
    x = Node(val = 1.5)

    l = rv.log(x) #NOTE x MUST BE > 0
    l.grad_value = 1.0
    print("∂l/∂x = {}".format(x.grad())) # ∂l/∂x = 0.6666

    # EXPONENT TESTS
    x = Node(val = 1.5)

    m = rv.exp(x) 
    m.grad_value = 1.0
    print("∂m/∂x = {}".format(x.grad())) # ∂m/∂x = 4.48

    # SIN TESTS
    x = Node(val = 1.5)

    n = rv.sin(x) 
    n.grad_value = 1.0
    print("∂n/∂x = {}".format(x.grad())) # ∂n/∂x = 0.07

    # COS TESTS
    x = Node(val = 1.5)

    o = rv.cos(x) 
    o.grad_value = 1.0
    print("∂o/∂x = {}".format(x.grad())) # ∂o/∂x = -0.997

    # TAN TESTS
    x = Node(val = 1.5)

    p = rv.tan(x) 
    p.grad_value = 1.0
    print("∂p/∂x = {}".format(x.grad())) # ∂p/∂x = 199.85